(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;

/* Package-scope variables */
var SandstormDb, Packages, DevApps, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern, addStaticAsset, userPictureUrl, fetchPicture;

(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/db.js                                                                        //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
// This file defines the database schema.                                                             // 17
                                                                                                      // 18
// Users = new Mongo.Collection("users");                                                             // 19
// The users collection is special and can be accessed through `Meteor.users`.                        // 20
// See https://docs.meteor.com/#/full/meteor_users. Entries in the users collection correspond to     // 21
// accounts. Each represents a capability store belonging to a human.                                 // 22
//                                                                                                    // 23
// Note that accounts are not the same thing as identities. An account may have multiple identities   // 24
// attached to it. Identities provide a way to authenticate the user and are also used to present     // 25
// globally unique and stable user indentifiers to grains via `Grain.UserInfo.userId`.                // 26
//                                                                                                    // 27
// Each entry in this collection contains:                                                            // 28
//   _id: Random.                                                                                     // 29
//   createdAt: Date when this entry was added to the collection.                                     // 30
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.               // 31
//   profile: Object containing profile data editable by the user and visible to other users. May     // 32
//            include the following fields. Note that if any field is missing, the first fallback     // 33
//            is to check `services` for details provided by the identity provider (the details       // 34
//            of which differ per-provider). Only if that is also missing do we fall back to          // 35
//            defaults.                                                                               // 36
//       name: String containing the display name of the user. Default: first part of email.          // 37
//       handle: String containing the user's preferred handle. Default: first part of email.         // 38
//       picture: _id into the StaticAssets table for the user's picture. Default: identicon.         // 39
//       pronoun: One of "male", "female", "neutral", or "robot". Default: neutral.                   // 40
//   services: Object containing login and identity data used by Meteor authentication services.      // 41
//   isAdmin: Boolean indicating whether this user is allowed to access the Sandstorm admin panel.    // 42
//   signupKey: If this is an invited user, then this field contains their signup key.                // 43
//   signupNote: If the user was invited through a link, then this field contains the note that the   // 44
//               inviter admin attached to the key.                                                   // 45
//   signupEmail: If the user was invited by email, then this field contains the email address that   // 46
//                the invite was sent to.                                                             // 47
//   plan: _id of an entry in the Plans table which determines the user's qutoa.                      // 48
//   storageUsage: Number of bytes this user is currently storing.                                    // 49
//   expires: Date when this user's account should be deleted. Only present for demo users.           // 50
//   isAppDemoUser: True if this is a demo user who arrived via an /appdemo/ link.                    // 51
//   payments: Object defined by payments module, if loaded.                                          // 52
                                                                                                      // 53
Packages = new Mongo.Collection("packages");                                                          // 54
// Packages which are installed or downloading.                                                       // 55
//                                                                                                    // 56
// Each contains:                                                                                     // 57
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                   // 58
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"  // 59
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),               // 60
//       >1 = download byte count.                                                                    // 61
//   error:  If status is "failed", error message string.                                             // 62
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.         // 63
//   appId:  If status is "ready", the application ID string.  Packages representing different        // 64
//       versions of the same app have the same appId.  The spk tool defines the app ID format        // 65
//       and can cryptographically verify that a package belongs to a particular app ID.              // 66
                                                                                                      // 67
DevApps = new Mongo.Collection("devapps");                                                            // 68
// List of applications currently made available via the dev tools running on the local machine.      // 69
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool    // 70
// on the local machine to publish an under-development app to this server. That should only ever     // 71
// happen on developers' desktop machines.                                                            // 72
//                                                                                                    // 73
// While a dev app is published, it automatically appears as installed by every user of the server,   // 74
// and it overrides all packages with the same application ID. If any instances of those packages     // 75
// are currently open, they are killed and reset on publish.                                          // 76
//                                                                                                    // 77
// When the dev tool disconnects, the app is automatically unpublished, and any open instances        // 78
// are again killed and refreshed.                                                                    // 79
//                                                                                                    // 80
// Each contains:                                                                                     // 81
//   _id:  The application ID string (as with Packages.appId).                                        // 82
//   packageId:  The directory name where the dev package is mounted.                                 // 83
//   timestamp:  Time when the package was last updated. If this changes while the package is         // 84
//     published, all running instances are reset. This is used e.g. to reset the app each time       // 85
//     changes are made to the source code.                                                           // 86
//   manifest:  The app's manifest, as with Packages.manifest.                                        // 87
                                                                                                      // 88
UserActions = new Mongo.Collection("userActions");                                                    // 89
// List of actions that each user has installed which create new grains.  Each app may install        // 90
// some number of actions (usually, one).                                                             // 91
//                                                                                                    // 92
// Each contains:                                                                                     // 93
//   _id:  random                                                                                     // 94
//   userId:  User who has installed this action.                                                     // 95
//   packageId:  Package used to run this action.                                                     // 96
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 97
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle.defaultText; denormalized so    // 98
//       that clients can access it without subscribing to the Packages collection.                   // 99
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 100
//       searchability.                                                                               // 101
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"              // 102
//   title:  Human-readable title for this action, e.g. "New Spreadsheet".                            // 103
//   command:  Manifest.Command to run this action (see package.capnp).                               // 104
                                                                                                      // 105
Grains = new Mongo.Collection("grains");                                                              // 106
// Grains belonging to users.                                                                         // 107
//                                                                                                    // 108
// Each contains:                                                                                     // 109
//   _id:  random                                                                                     // 110
//   packageId:  _id of the package of which this grain is an instance.                               // 111
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 112
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 113
//       searchability.                                                                               // 114
//   userId:  User who owns this grain.                                                               // 115
//   title:  Human-readable string title, as chosen by the user.                                      // 116
//   lastUsed:  Date when the grain was last used by a user.                                          // 117
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                   // 118
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent   // 119
//                   time a session to this grain was opened.                                         // 120
//                                                                                                    // 121
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and    // 122
// web publishing functionality without powerbox support; they will be replaced once the powerbox     // 123
// is implemented.                                                                                    // 124
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and    // 125
//       web publishing. This field is initialized when first requested by the app.                   // 126
                                                                                                      // 127
RoleAssignments = new Mongo.Collection("roleAssignments");                                            // 128
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges   // 129
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens            // 130
// collection.                                                                                        // 131
                                                                                                      // 132
Contacts = new Mongo.Collection("contacts");                                                          // 133
// Edges in the social graph.                                                                         // 134
//                                                                                                    // 135
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob        // 136
// will show up in her user-picker UI for actions like share-by-identity.                             // 137
//                                                                                                    // 138
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of       // 139
// Bob's.                                                                                             // 140
//                                                                                                    // 141
// Each contains:                                                                                     // 142
//   _id: random                                                                                      // 143
//   ownerId: The `_id` of the user who owns this contact.                                            // 144
//   userId:  The `_id` of the contacted user.                                                        // 145
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies       // 146
//            the contact to the owner.                                                               // 147
//   created: Date when this contact was created.                                                     // 148
                                                                                                      // 149
Sessions = new Mongo.Collection("sessions");                                                          // 150
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.   // 151
//                                                                                                    // 152
// Each contains:                                                                                     // 153
//   _id:  random                                                                                     // 154
//   grainId:  _id of the grain to which this session is connected.                                   // 155
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the    // 156
//       '*' in WILDCARD_HOST.                                                                        // 157
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some       // 158
//       period.                                                                                      // 159
//   userId:  User who owns this session.                                                             // 160
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens     // 161
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),        // 162
//       anonymous users can get access without an API token and so neither userId nor hashedToken    // 163
//       are present.                                                                                 // 164
//   powerboxView: If present, this is a view that should be presented as part of a powerbox          // 165
//       interaction.                                                                                 // 166
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                  // 167
                                                                                                      // 168
SignupKeys = new Mongo.Collection("signupKeys");                                                      // 169
// Invite keys which may be used by users to get access to Sandstorm.                                 // 170
//                                                                                                    // 171
// Each contains:                                                                                     // 172
//   _id:  random                                                                                     // 173
//   used:  Boolean indicating whether this key has already been consumed.                            // 174
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.         // 175
//   email: If this key was sent as an email invite, the email address to which it was sent.          // 176
                                                                                                      // 177
ActivityStats = new Mongo.Collection("activityStats");                                                // 178
// Contains usage statistics taken on a regular interval. Each entry is a data point.                 // 179
//                                                                                                    // 180
// Each contains:                                                                                     // 181
//   timestamp: Date when measurements were taken.                                                    // 182
//   daily: Contains stats counts pertaining to the last day before the sample time.                  // 183
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.          // 184
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.          // 185
//                                                                                                    // 186
// Each of daily, weekly, and monthly contains:                                                       // 187
//   activeUsers: The number of unique users who have used a grain on the server in the time          // 188
//       interval. Only counts logged-in users.                                                       // 189
//   demoUsers: Demo users.                                                                           // 190
//   appDemoUsers: Users that came in through "app demo".                                             // 191
//   activeGrains: The number of unique grains that have been used in the time interval.              // 192
                                                                                                      // 193
DeleteStats = new Mongo.Collection("deleteStats");                                                    // 194
// Contains records of objects that were deleted, for stat-keeping purposes.                          // 195
//                                                                                                    // 196
// Each contains:                                                                                     // 197
//   type: "grain" or "user" or "demoUser" or "appDemoUser"                                           // 198
//   lastActive: Date of the user's or grain's last activity.                                         // 199
                                                                                                      // 200
FileTokens = new Mongo.Collection("fileTokens");                                                      // 201
// Tokens corresponding to backup files that are currently stored on the server. A user receives      // 202
// a token when they create a backup file (either by uploading it, or by backing up one of their      // 203
// grains) and may use the token to read the file (either to download it, or to restore a new         // 204
// grain from it).                                                                                    // 205
//                                                                                                    // 206
// Each contains:                                                                                     // 207
//   _id:       The unguessable token string.                                                         // 208
//   name:      Suggested filename.                                                                   // 209
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.       // 210
                                                                                                      // 211
ApiTokens = new Mongo.Collection("apiTokens");                                                        // 212
// Access tokens for APIs exported by apps.                                                           // 213
//                                                                                                    // 214
// Originally API tokens were only used by external users through the HTTP API endpoint. However,     // 215
// now they are also used to implement SturdyRefs, not just held by external users, but also when     // 216
// an app holds a SturdyRef to another app within the same server. See the various `save()`,          // 217
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and        // 218
// `MainView`) -- the fields of type `Data` are API tokens.                                           // 219
//                                                                                                    // 220
// Each contains:                                                                                     // 221
//   _id:       A SHA-256 hash of the token.                                                          // 222
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)     // 223
//   userId:    For UiView capabilities, this is the user for whom the view is attenuated. That       // 224
//              is, the UiView's newSession() method will intersect the requested permissions with    // 225
//              this user's permissions before forwarding on to the underlying app. If `userId` is    // 226
//              not present, then no user attenuation is applied, i.e. this is a raw UiView as        // 227
//              implemented by the app. (The `roleAssignment` field, below, may still apply.)         // 228
//              For non-UiView capabilities, `userId` is never present.                               // 229
//              Note that this is NOT the user against whom the `requiredPermissions` parameter of    // 230
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerUser`.      // 231
//   userInfo:  *DEPRECATED* For API tokens created by the app through HackSessionContext, the        // 232
//              UserInfo struct that should be passed to `newSession()` when exercising this token,   // 233
//              in decoded (JS object) format. This is a temporary hack. `userId` is never present    // 234
//              when `userInfo` is present.                                                           // 235
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded        // 236
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These   // 237
//              permissions will be intersected with those held by `userId` when the view is opened.  // 238
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as   // 239
//              anonymous rather than as directly associated with `userId`. This has no effect on     // 240
//              the permissions granted.                                                              // 241
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by     // 242
//              the app or its supervisor (whereas without this it strictly represents UiView).       // 243
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).     // 244
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is            // 245
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.              // 246
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,   // 247
//              not a particular grain. (`grainId` and `userId` are not set.) This is an object       // 248
//              containing exactly one of the following fields:                                       // 249
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                   // 250
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the          // 251
//                           `Notifications` collection.                                              // 252
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this     // 253
//                  will be moved out of the frontend and into the backend, but we'll migrate the     // 254
//                  database when that happens. This field contains the boolean true to signify that  // 255
//                  it has been set.                                                                  // 256
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                    // 257
//   parentToken: If present, then this token represents exactly the capability represented by        // 258
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated   // 259
//              by `roleAssignment` (if present). To facilitate permissions computations, if the      // 260
//              capability is a UiView, then `grainId` is set to the backing grain and `userId` is    // 261
//              set to the user who shared the view. None of `userInfo`, `objectId`, or               // 262
//              `frontendRef` are present when `parentToken` is present.                              // 263
//   petname:   Human-readable label for this access token, useful for identifying tokens for         // 264
//              revocation. This should be displayed when visualizing incoming capabilities to        // 265
//              the grain identified by `grainId`.                                                    // 266
//   created:   Date when this token was created.                                                     // 267
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may   // 268
//              become un-revoked in the future.                                                      // 269
//   expires:   Optional expiration Date. If undefined, the token does not expire.                    // 270
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)         // 271
//              as passed to the `save()` call that created this token. If not present, treat         // 272
//              as `webkey` (the default for `ApiTokenOwner`).                                        // 273
//   expiresIfUnused:                                                                                 // 274
//              Optional Date after which the token, if it has not been used yet, expires.            // 275
//              This field should be cleared on a token's first use.                                  // 276
//   requirements: List of conditions which must hold for this token to be considered valid.          // 277
//              Semantically, this list specifies the powers which were *used* to originally          // 278
//              create the token. If any condition in the list becomes untrue, then the token must    // 279
//              be considered revoked, and all live refs and sturdy refs obtained transitively        // 280
//              through it must also become revoked. Each item is the JSON serialization of the       // 281
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                        // 282
//                                                                                                    // 283
// It is important to note that a token's owner and provider are independent from each other. To      // 284
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:   // 285
//                                                                                                    // 286
// struct ApiToken {                                                                                  // 287
//   owner :ApiTokenOwner;                                                                            // 288
//   provider :union {                                                                                // 289
//     grain :group {                                                                                 // 290
//       grainId :Text;                                                                               // 291
//       union {                                                                                      // 292
//         uiView :group {                                                                            // 293
//           userId :Text;                                                                            // 294
//           roleAssignment :RoleAssignment;                                                          // 295
//           forSharing :Bool;                                                                        // 296
//         }                                                                                          // 297
//         objectId :SupervisorObjectId;                                                              // 298
//       }                                                                                            // 299
//     }                                                                                              // 300
//     frontendRef :union {                                                                           // 301
//        notificationHandle :Text;                                                                   // 302
//        ipNetwork :Bool;                                                                            // 303
//        ipInterface :Bool;                                                                          // 304
//     }                                                                                              // 305
//     child :group {                                                                                 // 306
//       parentToken :Text;                                                                           // 307
//       union {                                                                                      // 308
//         uiView :group {                                                                            // 309
//           grainId :Text;                                                                           // 310
//           userId :Text;                                                                            // 311
//           roleAssignment :RoleAssignment = (allAccess = ());                                       // 312
//         }                                                                                          // 313
//         other :Void;                                                                               // 314
//       }                                                                                            // 315
//     }                                                                                              // 316
//   }                                                                                                // 317
//   requirements: List(Supervisor.MembraneRequirement);                                              // 318
//   ...                                                                                              // 319
// }                                                                                                  // 320
                                                                                                      // 321
Notifications = new Mongo.Collection("notifications");                                                // 322
// Notifications for a user.                                                                          // 323
//                                                                                                    // 324
// Each contains:                                                                                     // 325
//   _id:          random                                                                             // 326
//   ongoing:      If present, this is an ongoing notification, and this field contains an            // 327
//                 ApiToken referencing the `OngoingNotification` capability.                         // 328
//   grainId:      The grain originating this notification, if any.                                   // 329
//   userId:       The user receiving the notification.                                               // 330
//   text:         The JSON-ified LocalizedText to display in the notification.                       // 331
//   isUnread:     Boolean indicating if this notification is unread.                                 // 332
//   timestamp:    Date when this notification was last updated                                       // 333
                                                                                                      // 334
StatsTokens = new Mongo.Collection("statsTokens");                                                    // 335
// Access tokens for the Stats collection                                                             // 336
//                                                                                                    // 337
// These tokens are used for accessing the ActivityStats collection remotely                          // 338
// (ie. from a dashboard webapp)                                                                      // 339
//                                                                                                    // 340
// Each contains:                                                                                     // 341
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                 // 342
                                                                                                      // 343
Misc = new Mongo.Collection("misc");                                                                  // 344
// Miscellaneous configuration and other settings                                                     // 345
//                                                                                                    // 346
// This table is currently only used for persisting BASE_URL from one session to the next,            // 347
// but in general any miscellaneous settings should go in here                                        // 348
//                                                                                                    // 349
// Each contains:                                                                                     // 350
//   _id:       The name of the setting. eg. "BASE_URL"                                               // 351
//   value:     The value of the setting.                                                             // 352
                                                                                                      // 353
Settings = new Mongo.Collection("settings");                                                          // 354
// Settings for this Sandstorm instance go here. They are configured through the adminSettings        // 355
// route. This collection differs from misc in that this collection is completely user controlled.    // 356
//                                                                                                    // 357
// Each contains:                                                                                     // 358
//   _id:       The name of the setting. eg. "MAIL_URL"                                               // 359
//   value:     The value of the setting.                                                             // 360
//   potentially other fields that are unique to the setting                                          // 361
                                                                                                      // 362
Migrations = new Mongo.Collection("migrations");                                                      // 363
// This table tracks which migrations we have applied to this instance.                               // 364
// It contains a single entry:                                                                        // 365
//   _id:       "migrations_applied"                                                                  // 366
//   value:     The number of migrations this instance has successfully completed.                    // 367
                                                                                                      // 368
StaticAssets = new Mongo.Collection("staticAssets");                                                  // 369
// Collection of static assets served up from the Sandstorm server's "static" host. We only           // 370
// support relatively small assets: under 1MB each.                                                   // 371
//                                                                                                    // 372
// Each contains:                                                                                     // 373
//   _id:       Random ID; will be used in the URL.                                                   // 374
//   hash:      A SHA-256 hash of the data, used to de-dupe.                                          // 375
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                             // 376
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                   // 377
//   content:   The asset content (byte buffer).                                                      // 378
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't   // 379
//       have transactions, this needs to bias towards over-counting; a backup GC could be used       // 380
//       to catch leaked assets, although it's probably not a big deal in practice.                   // 381
                                                                                                      // 382
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                        // 383
// Collection of tokens representing a single-use permission to upload an asset, such as a new        // 384
// profile picture.                                                                                   // 385
//                                                                                                    // 386
// Each contains:                                                                                     // 387
//   _id:       Random ID.                                                                            // 388
//   purpose:   Contains one of the following, indicating how the asset is to be used:                // 389
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:         // 390
//           userId: User whose picture shall be replaced.                                            // 391
//           identityId: Which of the user's identities shall be updated.                             // 392
//   expires:   Time when this token will go away if unused.                                          // 393
                                                                                                      // 394
Plans = new Mongo.Collection("plans");                                                                // 395
// Subscription plans, which determine quota.                                                         // 396
//                                                                                                    // 397
// Each contains:                                                                                     // 398
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...               // 399
//   storage: Number of bytes this user is allowed to store.                                          // 400
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                         // 401
//   computeLabel: Label to display to the user describing this plan's compute units.                 // 402
//   grains: Total number of grains this user can create (often `Infinity`).                          // 403
//   price: Price per month in US cents.                                                              // 404
                                                                                                      // 405
if (Meteor.isServer) {                                                                                // 406
  Meteor.publish("credentials", function () {                                                         // 407
    // Data needed for isSignedUp() and isAdmin() to work.                                            // 408
                                                                                                      // 409
    if (this.userId) {                                                                                // 410
      return [                                                                                        // 411
        Meteor.users.find({_id: this.userId},                                                         // 412
            {fields: {signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,                          // 413
                      plan: 1, hasCompletedSignup: 1}}),                                              // 414
        Plans.find()                                                                                  // 415
      ];                                                                                              // 416
    } else {                                                                                          // 417
      return [];                                                                                      // 418
    }                                                                                                 // 419
  });                                                                                                 // 420
}                                                                                                     // 421
                                                                                                      // 422
isDemoUser = function() {                                                                             // 423
  // Returns true if this is a demo user.                                                             // 424
                                                                                                      // 425
  var user = Meteor.user();                                                                           // 426
  if (user && user.expires) {                                                                         // 427
    return true;                                                                                      // 428
  } else {                                                                                            // 429
    return false;                                                                                     // 430
  }                                                                                                   // 431
}                                                                                                     // 432
                                                                                                      // 433
isSignedUp = function() {                                                                             // 434
  // Returns true if the user has presented an invite key.                                            // 435
                                                                                                      // 436
  var user = Meteor.user();                                                                           // 437
                                                                                                      // 438
  if (!user) return false;  // not signed in                                                          // 439
                                                                                                      // 440
  if (user.expires) return false;  // demo user.                                                      // 441
                                                                                                      // 442
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                    // 443
                                                                                                      // 444
  if (user.signupKey) return true;  // user is invited                                                // 445
                                                                                                      // 446
  return false;                                                                                       // 447
}                                                                                                     // 448
                                                                                                      // 449
isSignedUpOrDemo = function () {                                                                      // 450
  var user = Meteor.user();                                                                           // 451
                                                                                                      // 452
  if (!user) return false;  // not signed in                                                          // 453
                                                                                                      // 454
  if (user.expires) return true;  // demo user.                                                       // 455
                                                                                                      // 456
  if (Meteor.settings.public.allowUninvited) return true;  // all accounts qualify                    // 457
                                                                                                      // 458
  if (user.signupKey) return true;  // user is invited                                                // 459
                                                                                                      // 460
  return false;                                                                                       // 461
}                                                                                                     // 462
                                                                                                      // 463
isUserOverQuota = function (user) {                                                                   // 464
  // Return false if user has quota space remaining, true if it is full. When this returns true,      // 465
  // we will not allow the user to create new grains, though they may be able to open existing ones   // 466
  // which may still increase their storage usage.                                                    // 467
  //                                                                                                  // 468
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                 // 469
                                                                                                      // 470
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                             // 471
                                                                                                      // 472
  var plan = Plans.findOne(user.plan || "free");                                                      // 473
                                                                                                      // 474
  if (plan.grains < Infinity) {                                                                       // 475
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains}).count();            // 476
    if (count >= plan.grains) return "outOfGrains";                                                   // 477
  }                                                                                                   // 478
                                                                                                      // 479
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";            // 480
}                                                                                                     // 481
                                                                                                      // 482
isUserExcessivelyOverQuota = function (user) {                                                        // 483
  // Return true if user is so far over quota that we should prevent their existing grains from       // 484
  // running at all.                                                                                  // 485
  //                                                                                                  // 486
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                 // 487
                                                                                                      // 488
  if (!Meteor.settings.public.quotaEnabled || user.isAdmin) return false;                             // 489
                                                                                                      // 490
  var plan = Plans.findOne(user.plan || "free");                                                      // 491
                                                                                                      // 492
  if (plan.grains < Infinity) {                                                                       // 493
    var count = Grains.find({userId: user._id}, {fields: {}, limit: plan.grains * 2}).count();        // 494
    if (count >= plan.grains * 2) return "outOfGrains";                                               // 495
  }                                                                                                   // 496
                                                                                                      // 497
  return plan && user.storageUsage && user.storageUsage >= plan.storage * 1.2 && "outOfStorage";      // 498
}                                                                                                     // 499
                                                                                                      // 500
isAdmin = function() {                                                                                // 501
  // Returns true if the user is the administrator.                                                   // 502
                                                                                                      // 503
  var user = Meteor.user();                                                                           // 504
  if (user && user.isAdmin) {                                                                         // 505
    return true;                                                                                      // 506
  } else {                                                                                            // 507
    return false;                                                                                     // 508
  }                                                                                                   // 509
}                                                                                                     // 510
                                                                                                      // 511
isAdminById = function(id) {                                                                          // 512
  // Returns true if the user's id is the administrator.                                              // 513
                                                                                                      // 514
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                  // 515
  if (user && user.isAdmin) {                                                                         // 516
    return true;                                                                                      // 517
  } else {                                                                                            // 518
    return false;                                                                                     // 519
  }                                                                                                   // 520
}                                                                                                     // 521
                                                                                                      // 522
findAdminUserForToken = function (token) {                                                            // 523
  if (!token.requirements) {                                                                          // 524
    return;                                                                                           // 525
  }                                                                                                   // 526
  var requirements = token.requirements.filter(function (requirement) {                               // 527
    return "userIsAdmin" in requirement;                                                              // 528
  });                                                                                                 // 529
                                                                                                      // 530
  if (requirements.length > 1) {                                                                      // 531
    return;                                                                                           // 532
  }                                                                                                   // 533
  if (requirements.length === 0) {                                                                    // 534
    return;                                                                                           // 535
  }                                                                                                   // 536
  return requirements[0].userIsAdmin;                                                                 // 537
};                                                                                                    // 538
                                                                                                      // 539
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                      // 540
                                                                                                      // 541
if (wildcardHost.length != 2) {                                                                       // 542
  throw new Error("Wildcard host must contain exactly one asterisk.");                                // 543
}                                                                                                     // 544
                                                                                                      // 545
matchWildcardHost = function(host) {                                                                  // 546
  // See if the hostname is a member of our wildcard. If so, extract the ID.                          // 547
                                                                                                      // 548
  var prefix = wildcardHost[0];                                                                       // 549
  var suffix = wildcardHost[1];                                                                       // 550
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                             // 551
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                    // 552
      host.length >= prefix.length + suffix.length) {                                                 // 553
    var id = host.slice(prefix.length, -suffix.length);                                               // 554
    if (id.match(/^[a-z0-9]*$/)) {                                                                    // 555
      return id;                                                                                      // 556
    }                                                                                                 // 557
  }                                                                                                   // 558
                                                                                                      // 559
  return null;                                                                                        // 560
}                                                                                                     // 561
                                                                                                      // 562
makeWildcardHost = function (id) {                                                                    // 563
  return wildcardHost[0] + id + wildcardHost[1];                                                      // 564
}                                                                                                     // 565
                                                                                                      // 566
if (Meteor.isServer) {                                                                                // 567
  var Url = Npm.require("url");                                                                       // 568
  getWildcardOrigin = function () {                                                                   // 569
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're       // 570
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),     // 571
    // an astrisk is only allowed as the first character and must be followed by a period. So we need // 572
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can    // 573
    // really do. We also add the protocol to the front (again, that's what CSP wants).               // 574
                                                                                                      // 575
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.  // 576
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                          // 577
                                                                                                      // 578
    var dotPos = wildcardHost[1].indexOf(".");                                                        // 579
    if (dotPos < 0) {                                                                                 // 580
      return protocol + "//*";                                                                        // 581
    } else {                                                                                          // 582
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                        // 583
    }                                                                                                 // 584
  }                                                                                                   // 585
}                                                                                                     // 586
                                                                                                      // 587
allowDevAccounts = function () {                                                                      // 588
  var setting = Settings.findOne({_id: "devAccounts"});                                               // 589
  if (setting) {                                                                                      // 590
    return setting.value;                                                                             // 591
  } else {                                                                                            // 592
    return Meteor.settings && Meteor.settings.public &&                                               // 593
           Meteor.settings.public.allowDevAccounts;                                                   // 594
  }                                                                                                   // 595
};                                                                                                    // 596
                                                                                                      // 597
roleAssignmentPattern = {                                                                             // 598
  none : Match.Optional(null),                                                                        // 599
  allAccess: Match.Optional(null),                                                                    // 600
  roleId: Match.Optional(Match.Integer),                                                              // 601
  addPermissions: Match.Optional([Boolean]),                                                          // 602
  removePermissions: Match.Optional([Boolean]),                                                       // 603
};                                                                                                    // 604
                                                                                                      // 605
SandstormDb = function () {                                                                           // 606
  this.collections = {                                                                                // 607
    // Direct access to underlying collections. DEPRECATED.                                           // 608
    //                                                                                                // 609
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove     // 610
    //   direct access to the collections.                                                            // 611
                                                                                                      // 612
    packages: Packages,                                                                               // 613
    devApps: DevApps,                                                                                 // 614
    userActions: UserActions,                                                                         // 615
    grains: Grains,                                                                                   // 616
    contacts: Contacts,                                                                               // 617
    sessions: Sessions,                                                                               // 618
    signupKeys: SignupKeys,                                                                           // 619
    activityStats: ActivityStats,                                                                     // 620
    deleteStats: DeleteStats,                                                                         // 621
    fileTokens: FileTokens,                                                                           // 622
    apiTokens: ApiTokens,                                                                             // 623
    notifications: Notifications,                                                                     // 624
    statsTokens: StatsTokens,                                                                         // 625
    misc: Misc,                                                                                       // 626
    settings: Settings,                                                                               // 627
                                                                                                      // 628
    // Intentionally omitted:                                                                         // 629
    // - Migrations, since it's used only within this package.                                        // 630
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.   // 631
  };                                                                                                  // 632
};                                                                                                    // 633
                                                                                                      // 634
// TODO(cleanup): These methods should not be defined freestanding and should use collection          // 635
//   objects created in SandstormDb's constructor rather than globals.                                // 636
                                                                                                      // 637
_.extend(SandstormDb.prototype, {                                                                     // 638
  isDemoUser: isDemoUser,                                                                             // 639
  isSignedUp: isSignedUp,                                                                             // 640
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                 // 641
  isUserOverQuota: isUserOverQuota,                                                                   // 642
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                             // 643
  isAdmin: isAdmin,                                                                                   // 644
  isAdminById: isAdminById,                                                                           // 645
  findAdminUserForToken: findAdminUserForToken,                                                       // 646
  matchWildcardHost: matchWildcardHost,                                                               // 647
  makeWildcardHost: makeWildcardHost,                                                                 // 648
  allowDevAccounts: allowDevAccounts,                                                                 // 649
  roleAssignmentPattern: roleAssignmentPattern,                                                       // 650
});                                                                                                   // 651
                                                                                                      // 652
if (Meteor.isServer) {                                                                                // 653
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                        // 654
}                                                                                                     // 655
                                                                                                      // 656
// =======================================================================================            // 657
// Below this point are newly-written or refactored functions.                                        // 658
                                                                                                      // 659
_.extend(SandstormDb.prototype, {                                                                     // 660
  // TODO(cleanup): These methods shouldn't take a raw mongo queries or "aggregations" as inputs.     // 661
  //   We want methods corresponding to each kind of query that is performed in practice. Letting     // 662
  //   the caller pass an arbitrary query and aggregations is problematic because:                    // 663
  //   - We can't type-check them to prevent Mongo injections.                                        // 664
  //   - It defeats the purpose of clearly seeing what kinds of queries we need to support, and       // 665
  //     therefore what kind of indexes we might need.                                                // 666
  //   - It doesn't make it any easier for us to substitute a non-Mongo database in the future.       // 667
  //   - If the caller is just going to pass in a match-all query, we may be forcing Mongo into       // 668
  //     a slower path by using $and.                                                                 // 669
  userGrains: function userGrains (user, query, aggregations) {                                       // 670
    var filteredQuery = { $and: [ { userId: user}, query ] };                                         // 671
    return this.collections.grains.find(filteredQuery, aggregations);                                 // 672
  },                                                                                                  // 673
                                                                                                      // 674
  currentUserGrains: function currentUserGrains (query, aggregations) {                               // 675
    return this.userGrains(Meteor.userId(), query, aggregations);                                     // 676
  },                                                                                                  // 677
                                                                                                      // 678
  userApiTokens: function userApiTokens (user) {                                                      // 679
    return this.collections.apiTokens.find({'owner.user.userId': user});                              // 680
  },                                                                                                  // 681
                                                                                                      // 682
  currentUserApiTokens: function currentUserApiTokens () {                                            // 683
    return this.userApiTokens(Meteor.userId());                                                       // 684
  },                                                                                                  // 685
                                                                                                      // 686
  userActions: function userActions (user, query, aggregations) {                                     // 687
    var filteredQuery = { $and: [ {userId: user}, query ] };                                          // 688
    return this.collections.userActions.find(filteredQuery, aggregations);                            // 689
  },                                                                                                  // 690
                                                                                                      // 691
  currentUserActions: function currentUserActions (query, aggregations) {                             // 692
    return this.userActions(Meteor.userId(), query, aggregations);                                    // 693
  },                                                                                                  // 694
                                                                                                      // 695
  getPlan: function (id) {                                                                            // 696
    check(id, String);                                                                                // 697
    var plan = Plans.findOne(id);                                                                     // 698
    if (!plan) {                                                                                      // 699
      throw new Error("no such plan: " + id);                                                         // 700
    }                                                                                                 // 701
    return plan;                                                                                      // 702
  },                                                                                                  // 703
                                                                                                      // 704
  listPlans: function () {                                                                            // 705
    return Plans.find({}, {sort: {price: 1}});                                                        // 706
  },                                                                                                  // 707
                                                                                                      // 708
  getMyPlan: function () {                                                                            // 709
    var user = Meteor.user();                                                                         // 710
    return user && Plans.findOne(user.plan || "free");                                                // 711
  },                                                                                                  // 712
                                                                                                      // 713
  getMyUsage: function (user) {                                                                       // 714
    user = user || Meteor.user();                                                                     // 715
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                              // 716
      if (Meteor.isClient) {                                                                          // 717
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly       // 718
        // not reactive.                                                                              // 719
        return user.pseudoUsage;                                                                      // 720
      } else {                                                                                        // 721
        return {                                                                                      // 722
          grains: Grains.find({userId: user._id}).count(),                                            // 723
          storage: user.storageUsage || 0,                                                            // 724
          compute: 0   // not tracked yet                                                             // 725
        };                                                                                            // 726
      }                                                                                               // 727
    } else {                                                                                          // 728
      return {grains: 0, storage: 0, compute: 0};                                                     // 729
    }                                                                                                 // 730
  },                                                                                                  // 731
                                                                                                      // 732
  isUninvitedFreeUser: function () {                                                                  // 733
    if (!Meteor.settings.public.allowUninvited) return false;                                         // 734
                                                                                                      // 735
    var user = Meteor.user();                                                                         // 736
    return user && !user.expires && (!user.plan || user.plan === "free");                             // 737
  },                                                                                                  // 738
                                                                                                      // 739
  getSetting: function (name) {                                                                       // 740
    var setting = Settings.findOne(name);                                                             // 741
    return setting && setting.value;                                                                  // 742
  },                                                                                                  // 743
});                                                                                                   // 744
                                                                                                      // 745
if (Meteor.isServer) {                                                                                // 746
  var Crypto = Npm.require("crypto");                                                                 // 747
  var ContentType = Npm.require("content-type");                                                      // 748
  var Zlib = Npm.require("zlib");                                                                     // 749
                                                                                                      // 750
  // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.           // 751
  var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                   // 752
                                                                                                      // 753
  var BufferSmallerThan = function (limit) {                                                          // 754
    return Match.Where(function (buf) {                                                               // 755
      check(buf, Buffer);                                                                             // 756
      return buf.length < limit;                                                                      // 757
    });                                                                                               // 758
  }                                                                                                   // 759
                                                                                                      // 760
  var DatabaseId = Match.Where(function (s) {                                                         // 761
    check(s, String);                                                                                 // 762
    return !!s.match(/^[a-zA-Z0-9_]+$/);                                                              // 763
  });                                                                                                 // 764
                                                                                                      // 765
  addStaticAsset = function (metadata, content) {                                                     // 766
    // Add a new static asset to the database. If `content` is a string rather than a buffer, it      // 767
    // will be automatically gzipped before storage; do not specify metadata.encoding in this case.   // 768
                                                                                                      // 769
    if (typeof content === "string" && !metadata.encoding) {                                          // 770
      content = gzipSync(new Buffer(content, "utf8"));                                                // 771
      metadata.encoding = "gzip";                                                                     // 772
    }                                                                                                 // 773
                                                                                                      // 774
    check(metadata, {                                                                                 // 775
      mimeType: String,                                                                               // 776
      encoding: Match.Optional("gzip")                                                                // 777
    });                                                                                               // 778
    check(content, BufferSmallerThan(1 << 20));                                                       // 779
                                                                                                      // 780
    // Validate content type.                                                                         // 781
    metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                     // 782
                                                                                                      // 783
    var hasher = Crypto.createHash("sha256");                                                         // 784
    hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                       // 785
    hasher.update(content);                                                                           // 786
    var hash = hasher.digest("base64");                                                               // 787
                                                                                                      // 788
    var existing = StaticAssets.findAndModify({                                                       // 789
      query: {hash: hash, refcount: {$gte: 1}},                                                       // 790
      update: {$inc: {refcount: 1}},                                                                  // 791
      fields: {_id: 1, refcount: 1},                                                                  // 792
    });                                                                                               // 793
    if (existing) {                                                                                   // 794
      return existing._id;                                                                            // 795
    }                                                                                                 // 796
                                                                                                      // 797
    return StaticAssets.insert(_.extend({                                                             // 798
      hash: hash,                                                                                     // 799
      content: content,                                                                               // 800
      refcount: 1                                                                                     // 801
    }, metadata));                                                                                    // 802
  }                                                                                                   // 803
                                                                                                      // 804
  SandstormDb.prototype.addStaticAsset = addStaticAsset;                                              // 805
                                                                                                      // 806
  SandstormDb.prototype.refStaticAsset = function (id) {                                              // 807
    // Increment the refcount on an existing static asset.                                            // 808
    //                                                                                                // 809
    // You must call this BEFORE adding the new reference to the DB, in case of failure between       // 810
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 811
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 812
                                                                                                      // 813
    check(id, String);                                                                                // 814
                                                                                                      // 815
    var existing = StaticAssets.findAndModify({                                                       // 816
      query: {hash: hash},                                                                            // 817
      update: {$inc: {refcount: 1}},                                                                  // 818
      fields: {_id: 1, refcount: 1},                                                                  // 819
    });                                                                                               // 820
    if (!existing) {                                                                                  // 821
      throw new Error("refStaticAsset() called on asset that doesn't exist");                         // 822
    }                                                                                                 // 823
  }                                                                                                   // 824
                                                                                                      // 825
  SandstormDb.prototype.unrefStaticAsset = function (id) {                                            // 826
    // Decrement refcount on a static asset and delete if it has reached zero.                        // 827
    //                                                                                                // 828
    // You must call this AFTER removing the reference from the DB, in case of failure between        // 829
    // the two calls. (This way, the failure case is a storage leak, which is probably not a big      // 830
    // deal and can be fixed by GC, rather than a mysteriously missing asset.)                        // 831
                                                                                                      // 832
    check(id, String);                                                                                // 833
                                                                                                      // 834
    var existing = StaticAssets.findAndModify({                                                       // 835
      query: {_id: id},                                                                               // 836
      update: {$inc: {refcount: -1}},                                                                 // 837
      fields: {_id: 1, refcount: 1},                                                                  // 838
      new: true,                                                                                      // 839
    });                                                                                               // 840
    if (!existing) {                                                                                  // 841
      console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);        // 842
    } else if (existing.refcount <= 0) {                                                              // 843
      StaticAssets.remove({_id: existing._id});                                                       // 844
    }                                                                                                 // 845
  }                                                                                                   // 846
                                                                                                      // 847
  SandstormDb.prototype.getStaticAsset = function (id) {                                              // 848
    // Get a static asset's mimeType, encoding, and raw content.                                      // 849
                                                                                                      // 850
    check(id, String);                                                                                // 851
                                                                                                      // 852
    var asset = StaticAssets.findOne(id, {fields: {_id: 0, mimeType: 1, encoding: 1, content: 1}});   // 853
    if (asset) {                                                                                      // 854
      // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy       // 855
      //   here.                                                                                      // 856
      asset.content = new Buffer(asset.content);                                                      // 857
    }                                                                                                 // 858
    return asset;                                                                                     // 859
  }                                                                                                   // 860
                                                                                                      // 861
  SandstormDb.prototype.newAssetUpload = function (purpose) {                                         // 862
    check(purpose, {profilePicture: {userId: DatabaseId, identityId: Match.Optional(DatabaseId)}});   // 863
                                                                                                      // 864
    return AssetUploadTokens.insert({                                                                 // 865
      purpose: purpose,                                                                               // 866
      expires: new Date(Date.now() + 300000),  // in 5 minutes                                        // 867
    });                                                                                               // 868
  }                                                                                                   // 869
                                                                                                      // 870
  SandstormDb.prototype.fulfillAssetUpload = function (id) {                                          // 871
    // Indicates that the given asset upload has completed. It will be removed and its purpose        // 872
    // returned. If no matching upload exists, returns undefined.                                     // 873
                                                                                                      // 874
    check(id, String);                                                                                // 875
                                                                                                      // 876
    var upload = AssetUploadTokens.findAndModify({                                                    // 877
      query: {_id: id},                                                                               // 878
      remove: true                                                                                    // 879
    });                                                                                               // 880
                                                                                                      // 881
    if (upload.expires.valueOf() < Date.now()) {                                                      // 882
      return undefined;  // already expired                                                           // 883
    } else {                                                                                          // 884
      return upload.purpose;                                                                          // 885
    }                                                                                                 // 886
  }                                                                                                   // 887
                                                                                                      // 888
  function cleanupExpiredAssetUploads() {                                                             // 889
    AssetUploadTokens.remove({expires: {$lt: Date.now()}});                                           // 890
  }                                                                                                   // 891
                                                                                                      // 892
  // Cleanup tokens every hour, with a random phase.                                                  // 893
  Meteor.setTimeout(function () {                                                                     // 894
    cleanupExpiredAssetUploads();                                                                     // 895
    Meteor.setInterval(function () {                                                                  // 896
      cleanupExpiredAssetUploads();                                                                   // 897
    }, 3600000);                                                                                      // 898
  }, Math.floor(Math.random() * 3600000));                                                            // 899
}                                                                                                     // 900
                                                                                                      // 901
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/profile.js                                                                   //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var makeIdenticon;                                                                                    // 17
var httpProtocol;                                                                                     // 18
var sha256;                                                                                           // 19
                                                                                                      // 20
if (Meteor.isServer) {                                                                                // 21
  Meteor.publish("accountIdentities", function () {                                                   // 22
    if (!Meteor.userId) return [];                                                                    // 23
                                                                                                      // 24
    return [                                                                                          // 25
      Meteor.users.find(Meteor.userId,                                                                // 26
        {fields: {                                                                                    // 27
          "profile":1,                                                                                // 28
          "devName":1,                                                                                // 29
          "expires":1,                                                                                // 30
                                                                                                      // 31
          "services.google.id":1,                                                                     // 32
          "services.google.email":1,                                                                  // 33
          "services.google.verified_email":1,                                                         // 34
          "services.google.name":1,                                                                   // 35
          "services.google.picture":1,                                                                // 36
          "services.google.gender":1,                                                                 // 37
                                                                                                      // 38
          "services.github.id":1,                                                                     // 39
          "services.github.email":1,                                                                  // 40
          "services.github.username":1,                                                               // 41
                                                                                                      // 42
          "services.emailToken.email":1                                                               // 43
        }})                                                                                           // 44
    ];                                                                                                // 45
  });                                                                                                 // 46
                                                                                                      // 47
  makeIdenticon = function (id) {                                                                     // 48
    // We only make identicons client-side.                                                           // 49
    return undefined;                                                                                 // 50
  }                                                                                                   // 51
                                                                                                      // 52
  var Url = Npm.require("url");                                                                       // 53
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                            // 54
                                                                                                      // 55
  var Crypto = Npm.require("crypto");                                                                 // 56
  sha256 = function (data) {                                                                          // 57
    return Crypto.createHash("sha256").update(data).digest("hex");                                    // 58
  };                                                                                                  // 59
} else {                                                                                              // 60
  var identiconCache = {};                                                                            // 61
                                                                                                      // 62
  makeIdenticon = function (id) {                                                                     // 63
    // Given a cryptographic hash as input, generate an identicon. We always use the user's           // 64
    // identity ID as the input hash (the same thing passed to apps as the user ID).                  // 65
                                                                                                      // 66
    // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so        // 67
    // that apps can themselves use identicon.js to produce consistent identicons. As it turns out    // 68
    // identicon.js doesn't use the second half of the hash even if we provide it, but slice it       // 69
    // anyway to be safe.                                                                             // 70
    id = id.slice(0, 32);                                                                             // 71
                                                                                                      // 72
    if (id in identiconCache) {                                                                       // 73
      return identiconCache[id];                                                                      // 74
    }                                                                                                 // 75
                                                                                                      // 76
    // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.      // 77
    var data = new Identicon(id, 512).toString();                                                     // 78
    var result = "data:image/png;base64," + data;                                                     // 79
    identiconCache[id] = result;                                                                      // 80
    return result;                                                                                    // 81
  }                                                                                                   // 82
                                                                                                      // 83
  httpProtocol = window.location.protocol;                                                            // 84
                                                                                                      // 85
  sha256 = function (data) {                                                                          // 86
    return CryptoJS.SHA256(data).toString();                                                          // 87
  };                                                                                                  // 88
}                                                                                                     // 89
                                                                                                      // 90
var GENDERS = {male: "male", female: "female", neutral: "neutral", robot: "robot"};                   // 91
                                                                                                      // 92
function staticAssetUrl(id, staticHost) {                                                             // 93
  if (id) {                                                                                           // 94
    return staticHost + "/" + id;                                                                     // 95
  } else {                                                                                            // 96
    return undefined;                                                                                 // 97
  }                                                                                                   // 98
}                                                                                                     // 99
                                                                                                      // 100
function filterHandle(handle) {                                                                       // 101
  // Convert disallowed letters into underscores, but no more than one underscore in a row.           // 102
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                              // 103
}                                                                                                     // 104
                                                                                                      // 105
function emailToHandle(email) {                                                                       // 106
  // Turn an email address into a handle.                                                             // 107
                                                                                                      // 108
  if (!email) return undefined;                                                                       // 109
                                                                                                      // 110
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters        // 111
  // on the same account.                                                                             // 112
  var parts = email.split("@");                                                                       // 113
  var base = filterHandle(parts[0].split("+")[0]);                                                    // 114
                                                                                                      // 115
  var domain = (parts[1]||"").split(".");                                                             // 116
  if (domain[domain.length - 1] === "name") {                                                         // 117
    // Oh, a .name domain. Let's use                                                                  // 118
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                // 119
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info",           // 120
                         "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"],          // 121
                         base)) {                                                                     // 122
    // This is probably an address at a vanity domain. Use the domain itself as the handle.           // 123
    base = filterHandle(domain[0]);                                                                   // 124
  }                                                                                                   // 125
                                                                                                      // 126
  return filterHandle(base);                                                                          // 127
}                                                                                                     // 128
                                                                                                      // 129
function identityId(service, id) {                                                                    // 130
  return sha256(service + ":" + id);                                                                  // 131
}                                                                                                     // 132
                                                                                                      // 133
function googleIdentity(user, staticHost) {                                                           // 134
  var google = (user.services && user.services.google) || {};                                         // 135
  var profile = user.profile || {};                                                                   // 136
  var id = identityId("google", google.id);                                                           // 137
                                                                                                      // 138
  return {                                                                                            // 139
    service: "google",                                                                                // 140
    id: id,                                                                                           // 141
    name: profile.name || google.name || "Name Unknown",                                              // 142
    email: profile.email || google.email,                                                             // 143
    handle: profile.handle || emailToHandle(google.email) ||                                          // 144
            filterHandle(profile.name || google.name) || "unknown",                                   // 145
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 146
    pronoun: profile.pronoun || GENDERS[google.gender] || "neutral",                                  // 147
  }                                                                                                   // 148
}                                                                                                     // 149
                                                                                                      // 150
function githubIdentity(user, staticHost) {                                                           // 151
  var github = (user.services && user.services.github) || {};                                         // 152
  var profile = user.profile || {};                                                                   // 153
  var id = identityId("github", github.id);                                                           // 154
                                                                                                      // 155
  return {                                                                                            // 156
    service: "github",                                                                                // 157
    id: id,                                                                                           // 158
    name: profile.name || github.username || "Name Unknown",                                          // 159
    email: profile.email || github.email,                                                             // 160
    handle: profile.handle || filterHandle(github.username) ||                                        // 161
            filterHandle(profile.name) || "unknown",                                                  // 162
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 163
    pronoun: profile.pronoun,                                                                         // 164
  }                                                                                                   // 165
}                                                                                                     // 166
                                                                                                      // 167
function emailIdentity(user, staticHost) {                                                            // 168
  var email = (user.services && user.services.emailToken) || {};                                      // 169
  var profile = user.profile || {};                                                                   // 170
  var id = identityId("email", email.email);                                                          // 171
                                                                                                      // 172
  return {                                                                                            // 173
    service: "email",                                                                                 // 174
    id: id,                                                                                           // 175
    name: profile.name || email.email.split("@")[0] || "Name Unknown",                                // 176
    email: profile.email || email.email,                                                              // 177
    handle: profile.handle || emailToHandle(email.email) ||                                           // 178
            filterHandle(profile.name) || "unknown",                                                  // 179
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 180
    pronoun: profile.pronoun,                                                                         // 181
  }                                                                                                   // 182
}                                                                                                     // 183
                                                                                                      // 184
function devIdentity(user, staticHost) {                                                              // 185
  var email = (user.services && user.services.emailToken) || {};                                      // 186
  var profile = user.profile || {};                                                                   // 187
  var name = user.devName.split(" ")[0].toLowerCase();                                                // 188
  var id = identityId("dev", user.devName);                                                           // 189
                                                                                                      // 190
  return {                                                                                            // 191
    service: "dev",                                                                                   // 192
    id: id,                                                                                           // 193
    name: profile.name || user.devName,                                                               // 194
    handle: profile.handle || filterHandle(name),                                                     // 195
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 196
    pronoun: profile.pronoun ||                                                                       // 197
             (_.contains(["alice", "carol", "eve"], name) ? "female" :                                // 198
              _.contains(["bob", "dave"], name) ? "male" : "neutral"),                                // 199
  }                                                                                                   // 200
}                                                                                                     // 201
                                                                                                      // 202
function demoIdentity(user, staticHost) {                                                             // 203
  var email = (user.services && user.services.emailToken) || {};                                      // 204
  var profile = user.profile || {};                                                                   // 205
  var id = identityId("demo", user._id);                                                              // 206
                                                                                                      // 207
  return {                                                                                            // 208
    service: "demo",                                                                                  // 209
    id: id,                                                                                           // 210
    name: profile.name || "Demo User",                                                                // 211
    handle: profile.handle || "demo",                                                                 // 212
    picture: staticAssetUrl(profile.picture, staticHost) || makeIdenticon(id),                        // 213
    pronoun: profile.pronoun || "neutral",                                                            // 214
  }                                                                                                   // 215
}                                                                                                     // 216
                                                                                                      // 217
SandstormDb.getVerifiedEmails = function (user) {                                                     // 218
  var result = [];                                                                                    // 219
                                                                                                      // 220
  var services = user.services || {};                                                                 // 221
  var google = services.google || {};                                                                 // 222
  if (google.email && google.verified_email) result.push(google.email);                               // 223
  if (services.emailToken) result.push(services.emailToken.email);                                    // 224
                                                                                                      // 225
  // TODO(soon): Verification of email addresses -- perhaps through asking the user to log in as      // 226
  //   the given identity?                                                                            // 227
                                                                                                      // 228
  return result;                                                                                      // 229
}                                                                                                     // 230
                                                                                                      // 231
SandstormDb.getUserIdentities = function (user) {                                                     // 232
  // Given a user object, return all of the user's identities.                                        // 233
  //                                                                                                  // 234
  // On the client, must be subscribed "accountIdentities" for the user.                              // 235
                                                                                                      // 236
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                  // 237
                                                                                                      // 238
  var result = [];                                                                                    // 239
  if (user && user.services) {                                                                        // 240
    if ("google" in user.services) {                                                                  // 241
      result.push(googleIdentity(user, staticHost));                                                  // 242
    }                                                                                                 // 243
    if ("github" in user.services) {                                                                  // 244
      result.push(githubIdentity(user, staticHost));                                                  // 245
    }                                                                                                 // 246
    if ("emailToken" in user.services) {                                                              // 247
      result.push(emailIdentity(user, staticHost));                                                   // 248
    }                                                                                                 // 249
    if ("devName" in user) {                                                                          // 250
      result.push(devIdentity(user, staticHost));                                                     // 251
    }                                                                                                 // 252
    if ("expires" in user) {                                                                          // 253
      result.push(demoIdentity(user, staticHost));                                                    // 254
    }                                                                                                 // 255
  }                                                                                                   // 256
  return result;                                                                                      // 257
}                                                                                                     // 258
                                                                                                      // 259
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/user.js                                                                      //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
var Future = Npm.require("fibers/future");                                                            // 17
                                                                                                      // 18
userPictureUrl = function (user) {                                                                    // 19
  if (user.services && !(user.profile && user.profile.picture)) {                                     // 20
    // Try to determine user's avatar URL from login service.                                         // 21
                                                                                                      // 22
    var google = user.services.google;                                                                // 23
    if (google && google.picture) {                                                                   // 24
      return google.picture;                                                                          // 25
    }                                                                                                 // 26
                                                                                                      // 27
    var github = user.services.github;                                                                // 28
    if (github && github.id) {                                                                        // 29
      return "https://avatars.githubusercontent.com/u/" + github.id;                                  // 30
    }                                                                                                 // 31
                                                                                                      // 32
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be     // 33
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and         // 34
    // Github are different because they are actually the identity providers, so they already know    // 35
    // the user logged in.                                                                            // 36
  }                                                                                                   // 37
}                                                                                                     // 38
                                                                                                      // 39
fetchPicture = function (url) {                                                                       // 40
  try {                                                                                               // 41
    var result = HTTP.get(url, {                                                                      // 42
      npmRequestOptions: { encoding: null },                                                          // 43
      timeout: 5000                                                                                   // 44
    });                                                                                               // 45
                                                                                                      // 46
    var metadata = {};                                                                                // 47
                                                                                                      // 48
    metadata.mimeType = result.headers["content-type"];                                               // 49
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 &&                                       // 50
        metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {                                      // 51
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                 // 52
    }                                                                                                 // 53
    var enc = result.headers["content-encoding"];                                                     // 54
    if (enc && enc !== "identity") {                                                                  // 55
      metadata.encoding = enc;                                                                        // 56
    }                                                                                                 // 57
                                                                                                      // 58
    return addStaticAsset(metadata, result.content);                                                  // 59
  } catch (err) {                                                                                     // 60
    console.error("failed to fetch user profile picture:", url, err.stack);                           // 61
  }                                                                                                   // 62
}                                                                                                     // 63
                                                                                                      // 64
var ValidHandle = Match.Where(function (handle) {                                                     // 65
  check(handle, String);                                                                              // 66
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                        // 67
});                                                                                                   // 68
                                                                                                      // 69
Accounts.onCreateUser(function (options, user) {                                                      // 70
  // The first non-dev user to sign in should be automatically upgraded to admin.                     // 71
  // Dev users are identified by having the devName field.                                            // 72
  if (Meteor.users.find({devName: {$exists: 0}}).count() === 0 && !user.devName) {                    // 73
    user.isAdmin = true;                                                                              // 74
    user.signupKey = "admin";                                                                         // 75
  }                                                                                                   // 76
                                                                                                      // 77
  // Check profile.                                                                                   // 78
  if (options.profile) {                                                                              // 79
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                              // 80
    check(options.profile, Match.ObjectIncluding({                                                    // 81
      name: Match.OneOf(null, Match.Optional(String)),                                                // 82
      handle: Match.Optional(ValidHandle),                                                            // 83
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot")),                     // 84
    }));                                                                                              // 85
                                                                                                      // 86
    user.profile = options.profile;                                                                   // 87
  } else {                                                                                            // 88
    user.profile = {};                                                                                // 89
  }                                                                                                   // 90
                                                                                                      // 91
  // Try downloading avatar.                                                                          // 92
  var url = userPictureUrl(user);                                                                     // 93
  if (url) {                                                                                          // 94
    var assetId = fetchPicture(url);                                                                  // 95
    if (assetId) {                                                                                    // 96
      user.profile.picture = assetId;                                                                 // 97
    }                                                                                                 // 98
  }                                                                                                   // 99
                                                                                                      // 100
  return user;                                                                                        // 101
});                                                                                                   // 102
                                                                                                      // 103
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                   // 1
//                                                                                                    // 2
// Because migrations can experience partial failure and likely have                                  // 3
// side-effects, we should be careful to make sure all migrations are                                 // 4
// idempotent and safe to accidentally run multiple times.                                            // 5
                                                                                                      // 6
var updateLoginStyleToRedirect = function() {                                                         // 7
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 8
  ["google", "github"].forEach(function(serviceName) {                                                // 9
    var config = configurations.findOne({service: serviceName});                                      // 10
    if (config && config.loginStyle !== "redirect") {                                                 // 11
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                // 12
    }                                                                                                 // 13
  });                                                                                                 // 14
};                                                                                                    // 15
                                                                                                      // 16
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                          // 17
  // In the before time, Google and Github login were enabled by default.                             // 18
  //                                                                                                  // 19
  // This actually didn't make much sense, required the first user to configure                       // 20
  // OAuth, and had some trust-the-first-user properties that weren't totally                         // 21
  // secure.                                                                                          // 22
  //                                                                                                  // 23
  // Now, we have admin-token, so we wish to disable all logins by default                            // 24
  // (since they need to be configured anyway) but since users may have never                         // 25
  // explicitly told Sandstorm that Google or Github login should be enabled,                         // 26
  // we can't just use the value in the Settings collection, since it might                           // 27
  // never have been set.                                                                             // 28
  //                                                                                                  // 29
  // Thus, if the service is configured but absent from Settings, we should                           // 30
  // explicitly enable it in Settings, and then the rest of the logic can just                        // 31
  // depend on what value is in Settings and default to false without breaking                        // 32
  // user installations.                                                                              // 33
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 34
  ["google", "github"].forEach(function(serviceName) {                                                // 35
    var config = configurations.findOne({service: serviceName});                                      // 36
    var serviceConfig = Settings.findOne({_id: serviceName});                                         // 37
    if (config && !serviceConfig) {                                                                   // 38
      // Only explicitly enable the login service if:                                                 // 39
      // 1) the service is already configured                                                         // 40
      // 2) there is no sandstorm configuration already present (the user was                         // 41
      //    using the previous default behavior).                                                     // 42
      Settings.insert({_id: serviceName, value: true});                                               // 43
    }                                                                                                 // 44
  });                                                                                                 // 45
};                                                                                                    // 46
                                                                                                      // 47
var denormalizeInviteInfo = function() {                                                              // 48
  // When a user is invited via a signup token, the `signupKey` field of their user table entry       // 49
  // has always been populated to indicate the key they used. This points into the SignupKeys table   // 50
  // which has more information about the key, namely a freeform note entered by the admin when       // 51
  // they created the key. In the case that the email invite form was used, the note has the form     // 52
  // "E-mail invite to <address>".                                                                    // 53
  //                                                                                                  // 54
  // Later, we decided it was useful to indicate in the users table visible to the admin              // 55
  // information about the invite terms. Namely, for email invites we want to show the address        // 56
  // and for others we want to show the note. To make this efficient, fields `signupNote` and         // 57
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing       // 58
  // from the SignupKeys table.                                                                       // 59
                                                                                                      // 60
  Meteor.users.find().forEach(function (user) {                                                       // 61
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {       // 62
      var signupInfo = SignupKeys.findOne(user.signupKey);                                            // 63
      if (signupInfo && signupInfo.note) {                                                            // 64
        var newFields = { signupNote: signupInfo.note };                                              // 65
                                                                                                      // 66
        var prefix = "E-mail invite to ";                                                             // 67
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                              // 68
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                               // 69
        }                                                                                             // 70
                                                                                                      // 71
        Meteor.users.update(user._id, {$set: newFields});                                             // 72
      }                                                                                               // 73
    }                                                                                                 // 74
  });                                                                                                 // 75
}                                                                                                     // 76
                                                                                                      // 77
function mergeRoleAssignmentsIntoApiTokens() {                                                        // 78
  RoleAssignments.find().forEach(function(roleAssignment) {                                           // 79
    ApiTokens.insert({                                                                                // 80
      grainId: roleAssignment.grainId,                                                                // 81
      userId: roleAssignment.sharer,                                                                  // 82
      roleAssignment: roleAssignment.roleAssignment,                                                  // 83
      petname: roleAssignment.petname,                                                                // 84
      created: roleAssignment.created,                                                                // 85
      owner: {user: {userId: roleAssignment.recipient,                                                // 86
                     title: roleAssignment.title}},                                                   // 87
    });                                                                                               // 88
  });                                                                                                 // 89
}                                                                                                     // 90
                                                                                                      // 91
function fixOasisStorageUsageStats() {                                                                // 92
  // Before storage quota enforcement was first implemented, there were some bugs with the way the    // 93
  // Oasis backend tracked storage usage, which were not initially noticed since the resulting        // 94
  // number wasn't acutally used for anything.                                                        // 95
  //                                                                                                  // 96
  // With the underlying bugs fixed, simply starting up each grain (supervisor only, not the app)     // 97
  // and then immediately shutting it down should cause the sizes to be recalculated correctly.       // 98
  //                                                                                                  // 99
  // This problem only applies to Oasis.                                                              // 100
                                                                                                      // 101
  if (Meteor.settings.public.quotaEnabled) {                                                          // 102
    var dummyCommand = {                                                                              // 103
      argv: ["foo", "bar"],                                                                           // 104
      environ: []                                                                                     // 105
    };                                                                                                // 106
                                                                                                      // 107
    Grains.find().forEach(function (grain) {                                                          // 108
      try {                                                                                           // 109
        console.log("refreshing grain:", grain._id);                                                  // 110
                                                                                                      // 111
        // We take a highly synchronous approach here so that the logging is understandable.          // 112
        var supervisor = waitPromise(sandstormBackend.startGrain(                                     // 113
            grain.userId, grain._id, grain.packageId, dummyCommand, false, false)).supervisor;        // 114
        try {                                                                                         // 115
          waitPromise(supervisor.shutdown());                                                         // 116
          throw new Error("shutdown() didn't disconnect!");                                           // 117
        } catch (err) {                                                                               // 118
          if (err.kjType !== "disconnected") throw err;                                               // 119
        }                                                                                             // 120
      } catch (err) {                                                                                 // 121
        console.log(err.stack);                                                                       // 122
      }                                                                                               // 123
    });                                                                                               // 124
  }                                                                                                   // 125
}                                                                                                     // 126
                                                                                                      // 127
function fetchProfilePictures() {                                                                     // 128
  Meteor.users.find({}).forEach(function (user) {                                                     // 129
    var url = userPictureUrl(user);                                                                   // 130
    if (url) {                                                                                        // 131
      console.log("Fetching user picture:", url);                                                     // 132
      var assetId = fetchPicture(url);                                                                // 133
      if (assetId) {                                                                                  // 134
        Meteor.users.update(user._id, {$set: {"profile.picture": assetId}});                          // 135
      }                                                                                               // 136
    }                                                                                                 // 137
  });                                                                                                 // 138
}                                                                                                     // 139
                                                                                                      // 140
function assignPlans() {                                                                              // 141
  if (Meteor.settings.public.quotaEnabled && SandstormDb.paymentsMigrationHook) {                     // 142
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                              // 143
  }                                                                                                   // 144
}                                                                                                     // 145
                                                                                                      // 146
// This must come after all the functions named within are defined.                                   // 147
// Only append to this list!  Do not modify or remove list entries;                                   // 148
// doing so is likely change the meaning and semantics of user databases.                             // 149
var MIGRATIONS = [                                                                                    // 150
  updateLoginStyleToRedirect,                                                                         // 151
  enableLegacyOAuthProvidersIfNotInSettings,                                                          // 152
  denormalizeInviteInfo,                                                                              // 153
  mergeRoleAssignmentsIntoApiTokens,                                                                  // 154
  fixOasisStorageUsageStats,                                                                          // 155
  fetchProfilePictures,                                                                               // 156
  assignPlans,                                                                                        // 157
];                                                                                                    // 158
                                                                                                      // 159
function migrateToLatest() {                                                                          // 160
  var applied = Migrations.findOne({_id: "migrations_applied"});                                      // 161
  var start;                                                                                          // 162
  if (!applied) {                                                                                     // 163
    // Migrations table is not yet seeded with a value.  This means it has                            // 164
    // applied 0 migrations.  Persist this.                                                           // 165
    Migrations.insert({_id: "migrations_applied", value: 0});                                         // 166
    start = 0;                                                                                        // 167
  } else {                                                                                            // 168
    start = applied.value;                                                                            // 169
  }                                                                                                   // 170
  console.log("Migrations applied: " + start + "/" + MIGRATIONS.length);                              // 171
                                                                                                      // 172
  for (var i = start ; i < MIGRATIONS.length ; i++) {                                                 // 173
    // Apply migration i, then record that migration i was successfully run.                          // 174
    console.log("Applying migration " + (i+1));                                                       // 175
    MIGRATIONS[i]();                                                                                  // 176
    Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                             // 177
    console.log("Applied migration " + (i+1));                                                        // 178
  }                                                                                                   // 179
}                                                                                                     // 180
                                                                                                      // 181
// Apply all migrations on startup.                                                                   // 182
Meteor.startup(migrateToLatest);                                                                      // 183
                                                                                                      // 184
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
